"""Get deails about the ssl cert of a hostname such as:

- When the cert will expire
- The issuer of the cert
- A full dict of the details of the cert

Notes
-----
- You should use an FQDN for any ssl_utilties functions, so something like https://www.google.ca becomes google.ca

Examples
--------
### Check when kieranwood.ca SSL cert expires
```
from sws.ssl_utlities import check_ssl_expiry

print(check_ssl_expiry('kieranwood.ca')) # prints: 'Oct  9 12:00:00 2020 GMT'
```

### Print the full cert details of kieranwood.ca
```
from pprint import pprint

from sws.ssl_utlities import check_ssl_expiry

pprint(get_ssl_cert('kieranwood.ca'))
# Prints:
'''
    {'OCSP': ('http://ocsp.digicert.com',),
    'caIssuers': ('http://cacerts.digicert.com/CloudFlareIncECCCA-2.crt',),
    'crlDistributionPoints': ('http://crl3.digicert.com/CloudFlareIncECCCA2.crl',
                            'http://crl4.digicert.com/CloudFlareIncECCCA2.crl'),
    'issuer': ((('countryName', 'US'),),
                (('stateOrProvinceName', 'CA'),),
                (('localityName', 'San Francisco'),),
                (('organizationName', 'CloudFlare, Inc.'),),
                (('commonName', 'CloudFlare Inc ECC CA-2'),)),
    'notAfter': 'Oct  9 12:00:00 2020 GMT',
    'notBefore': 'Jan  8 00:00:00 2020 GMT',
    'serialNumber': '06C5F428AF7D8439EB013216CB300B6A',
    'subject': ((('countryName', 'US'),),
                (('stateOrProvinceName', 'CA'),),
                (('localityName', 'San Francisco'),),
                (('organizationName', 'Cloudflare, Inc.'),),
                (('commonName', 'sni.cloudflaressl.com'),)),
    'subjectAltName': (('DNS', 'kieranwood.ca'),
                        ('DNS', 'sni.cloudflaressl.com'),
                        ('DNS', '*.kieranwood.ca')),
    'version': 3}
'''
```

### Get issuer details of existing hostname
```
from sws.ssl_utilities import get_ssl_issuer

get_ssl_issuer("kieranwood.ca") # Returns [('countryName', 'US'), ('organizationName', 'Cloudflare, Inc.'), ('commonName', 'Cloudflare Inc ECC CA-3')]
```

### Get issuer details of non-existing hostname
```
from sws.ssl_utilities import get_ssl_issuer

get_ssl_issuer("kieranwood.com") # Returns False
```
"""

# Internal Dependencies
import ssl                      # Used to get details about SSL certs
import socket                   # Used to make a request to get SSL cert
import logging                  # Used in logging for debug and info messages
from typing import Union        # Used to help provide more detailed type hints


def check_ssl_expiry(hostname: str) -> str:
    """Allows you to check the SSL expiry for a FQDN;
    More specifically it will return the notAfter for the SSL cert associated with the FQDN.

    Arguments
    ---------
    hostname : str
        A string of a FQDN (root URL with no protocol) for example 'kieranwood.ca'.

    Returns
    -------
    str:
        A the expiry of the domain name in MMM DD HH:MM:SS YYYY TZ format

    Raises
    ------
    ValueError:
        If hostname is not a valid domain

    Example
    -------
    Check when kieranwood.ca SSL cert expires
    ```
    from sws.ssl_utlities import check_ssl_expiry

    print(check_ssl_expiry('kieranwood.ca')) # prints: 'Oct  9 12:00:00 2020 GMT'
    ```
    """
    logging.info(f"check_ssl_expiry(hostname={hostname})")
    try:
        logging.info(f"Getting cert info for {hostname}")
        cert = get_ssl_cert(hostname)
    except socket.gaierror:
        raise ValueError(f"Unable to connect to {hostname}")

    expiry_date = cert["notAfter"]

    logging.info(f"exiting check_ssl_expiry() and returning {expiry_date}")
    return expiry_date


def get_ssl_cert(hostname: str) -> dict:
    """Returns all available SSL information, such as notAfter, commonName etc.

    Arguments
    ---------
    hostname : str
        A string of a FQDN (root URL with no protocol) for example 'kieranwood.ca'

    Raises
    ------
    ValueError:
        If hostname is not a valid domain

    Returns
    -------
    dict:
        The details about the SSL certificate

    Example
    -------
    Print the full cert details of kieranwood.ca
    ```
    from pprint import pprint

    from sws.ssl_utlities import check_ssl_expiry

    pprint(get_ssl_cert('kieranwood.ca'))
    # Prints:
    '''
        {'OCSP': ('http://ocsp.digicert.com',),
        'caIssuers': ('http://cacerts.digicert.com/CloudFlareIncECCCA-2.crt',),
        'crlDistributionPoints': ('http://crl3.digicert.com/CloudFlareIncECCCA2.crl',
                                'http://crl4.digicert.com/CloudFlareIncECCCA2.crl'),
        'issuer': ((('countryName', 'US'),),
                    (('stateOrProvinceName', 'CA'),),
                    (('localityName', 'San Francisco'),),
                    (('organizationName', 'CloudFlare, Inc.'),),
                    (('commonName', 'CloudFlare Inc ECC CA-2'),)),
        'notAfter': 'Oct  9 12:00:00 2020 GMT',
        'notBefore': 'Jan  8 00:00:00 2020 GMT',
        'serialNumber': '06C5F428AF7D8439EB013216CB300B6A',
        'subject': ((('countryName', 'US'),),
                    (('stateOrProvinceName', 'CA'),),
                    (('localityName', 'San Francisco'),),
                    (('organizationName', 'Cloudflare, Inc.'),),
                    (('commonName', 'sni.cloudflaressl.com'),)),
        'subjectAltName': (('DNS', 'kieranwood.ca'),
                            ('DNS', 'sni.cloudflaressl.com'),
                            ('DNS', '*.kieranwood.ca')),
        'version': 3}
    '''
    ```
    """
    logging.info(f"get_ssl_cert(hostname={hostname})")
    # Strip protocols from hostname if they exist
    if hostname.startswith("https://"):
        logging.info(f"Stripping https:// protocol from {hostname}")
        hostname = hostname.replace("https://", "")
    elif hostname.startswith("http://"):
        logging.info(f"Stripping http:// protocol from {hostname}")
        hostname = hostname.replace("http://", "")

    try:
        logging.info("Making SSL socket connection to retrieve info")
        context_socket = _get_ssl_socket(hostname)
        context_socket.connect((hostname, 443))  # Connects to the host over a socket
        cert = context_socket.getpeercert()  # Dictionary containing all the certificate information
    except socket.gaierror:
        raise ValueError(f"Unable to connect to {hostname}")

    logging.info(f"exiting get_ssl_cert() and returning {cert}")
    return cert


def get_ssl_issuer(hostname: str) -> Union[list, bool]:
    """Get's the details for the issuer of the hostname's SSL cert

    Parameters
    ----------
    hostname : str
        The hostname you want to get the issuer for

    Raises
    ------
    ValueError:
        If hostname is not a valid domain

    Returns
    -------
    list[tuple[str,str]] or False
        Returns a list of tuples with details about the SSL issuer,
        or False if there is no SSL cert for associated domain

    Examples
    --------
    Get details of existing hostname
    ```
    from sws.ssl_utilities import get_ssl_issuer

    get_ssl_issuer("kieranwood.ca") # Returns [('countryName', 'US'), ('organizationName', 'Cloudflare, Inc.'), ('commonName', 'Cloudflare Inc ECC CA-3')]
    ```

    Get details of non-existing hostname
    ```
    from sws.ssl_utilities import get_ssl_issuer

    get_ssl_issuer("kieranwood.com") # Returns False
    ```
    """
    logging.info(f"get_ssl_issuer(hostname={hostname})")
    # Strip protocols from hostname if they exist
    if hostname.startswith("https://"):
        logging.info(f"Stripping https:// protocol from {hostname}")
        hostname = hostname.replace("https://", "")
    elif hostname.startswith("http://"):
        logging.info(f"Stripping http:// protocol from {hostname}")
        hostname = hostname.replace("http://", "")
    try:
        cert = get_ssl_cert(hostname)
    except socket.gaierror:
        raise ValueError(f"Unable to connect to {hostname}")
    if cert["issuer"]:
        issuer = [data[0] for data in cert["issuer"]]
        logging.info(f"exiting get_ssl_issuer() and returning {issuer}")
        return issuer
    else:
        logging.info("exiting get_ssl_issuer() and returning False")
        return False


def _get_ssl_socket(hostname: str) -> ssl.SSLSocket:
    """Get's the context for a host over SSL

    Parameters
    ----------
    hostname: (str)
        The hostname to instantiate a SSL socket context to

    Raises
    ------
    ValueError:
        If hostname is not a valid domain

    Returns
    -------
    SSLSocket:
        A socket that can be used to connect to the hostname

    Examples
    --------
    ```
    context_socket = _get_ssl_context(hostname)

    context_socket.connect((hostname, 443)) # Connects to the host over a socket
    ```
    """
    # Strip protocols from hostname if they exist
    if hostname.startswith("http://"):
        hostname = hostname.replace("http://", "")
    elif hostname.startswith("https://"):
        hostname = hostname.replace("https://", "")
    context = ssl.create_default_context()
    context_socket = context.wrap_socket(socket.socket(), server_hostname=hostname)
    return context_socket



"""A module for dealing with getting domain information such as:

- If a domain is available
- Who a domain is registered with
- Other domain details such as, creation_date, name_servers etc.

Notes
-----
- If whois is not installed, the package/executable will be installed

Examples
--------
### Getting details about an available domain
```
from sws.domains import get_domain_info, domain_availability

domain_details = get_domain_info('kieranwood.com')
domain_availability(domain_details) # ('Domain available', True)
```

### Getting details of an unavailable domain
```
from sws.domains import get_domain_info, domain_availability

domain_details = get_domain_info('kieranwood.ca')
domain_availability(domain_details) # ('Domain unavailable until 6/November/2020', False)
```

### Getting details about an unregistered domain
```
from sws.domains import get_domain_info

print(get_domain_info('kieranwood.com')) # {'creation_date': False, 'expiration_date': False, 'last_updated': False, 'name': domain, 'name_servers': False, 'registrant_cc': False, 'registrar': False}
```

### Getting details of a registered domain
```
from sws.domains import get_domain_info

print(get_domain_info('kieranwood.ca')) # {'creation_date': datetime.datetime(2018, 11, 6, 5, 9, 47), 'expiration_date': datetime.datetime(2020, 11, 6, 5, 9, 47), 'last_updated': datetime.datetime(2020, 1, 8, 8, 9, 44), 'name': 'kieranwood.ca', 'name_servers': {'kevin.ns.cloudflare.com', 'sharon.ns.cloudflare.com'}, 'registrant_cc': 'redacted for privacy', 'registrar': 'Go Daddy Domains Canada, Inc'}
```
"""

# Standard Library Dependencies
import os                        # Used for path manipulation
import sys                       # Used to exit safely during errors
import logging                   # Used for logging in debugging etc.
import subprocess                # Used to execute existing binaries
from shutil import move          # Used to move folders within the os
from datetime import datetime    # Used for interpreting dates and times
from calendar import month_name  # Used to convert integer month representations to string representations

# Third Party Dependencies
import whois  # Used to pull domain information
from pystall.core import build, ZIPResource, _add_to_path, APTResource  # Used to install whois binary


def get_domain_info(domain: str) -> dict:
    """Returns a dictionary of all domain information

    Parameters
    ----------
    domain : str
        The domain you want the details for

    Returns
    -------
    dict
        A dictionary of a whois.Domain query

    Notes
    -----
    - If whois is not installed, the package/executable will be installed
    - Make sure to use the domain, and not just a url for example https://kieranwood.ca/hello is a url but kieranwood.ca is a domain
    - In the case that a protocol (http:// or https://) is provided it will be stripped, be aware this can cause comparison issues to the 'name' parameter of the dictionary
    
    Raises
    ------
    ValueError:
        If provided domain is not a valid domain (i.e. Subdomain, or URL instead of domain)

    Examples
    --------
    Getting details about an unregistered domain
    ```
    from sws.domains import get_domain_info

    print(get_domain_info('kieranwood.com')) # {'creation_date': False, 'expiration_date': False, 'last_updated': False, 'name': domain, 'name_servers': False, 'registrant_cc': False, 'registrar': False}
    ```

    Getting details of a registered domain
    ```
    from sws.domains import get_domain_info

    print(get_domain_info('kieranwood.ca')) # {'creation_date': datetime.datetime(2018, 11, 6, 5, 9, 47), 'expiration_date': datetime.datetime(2020, 11, 6, 5, 9, 47), 'last_updated': datetime.datetime(2020, 1, 8, 8, 9, 44), 'name': 'kieranwood.ca', 'name_servers': {'kevin.ns.cloudflare.com', 'sharon.ns.cloudflare.com'}, 'registrant_cc': 'redacted for privacy', 'registrar': 'Go Daddy Domains Canada, Inc'}
    ```
    """
    logging.info(f"Entering get_domain_info(domain={domain})")
    _install_whois()  # Verify/install whois

    # Validation

    ## Strip protocols
    if domain.startswith("https://"):
        logging.info(f"Stripping https:// protocol from {domain}")
        domain = domain.replace("https://", "")
    elif domain.startswith("http://"):
        logging.info(f"Stripping http:// protocol from {domain}")
        domain = domain.replace("http://", "")

    ## Raise error if subdomain
    logging.info(f"Confirming domain {domain} is not a subdomain")
    if len(domain.split(".")) > 2: # If domain is subdomain
        raise ValueError(f"Provided domain {domain} is likely a subdomain")
    
    ## Raise Error if invalid TLD
    try:
        logging.info(f"Querying {domain} with whois")
        domain_details = whois.query(domain)
    except Exception as e:
        if "Unknown TLD:" in str(e):
            raise ValueError(f"Domain {domain} is not a valid domain")

    # ## TODO: When new version of python-whois-extended releases uncomment below code
    # try:
        
    #     if os.name == "nt" and os.path.exists(os.path.realpath(f"{os.getenv('USERPROFILE')}\\..\\..\\whois")):
    #         INSTALL_FOLDER = os.path.realpath(f"{os.getenv('USERPROFILE')}\\..\\..\\whois\\whois.exe")
    #         domain_details = whois.query(domain, executable=INSTALL_FOLDER)
    #     else:
    #         domain_details = whois.query(domain)
    # except Exception as e:
    #     if "Unknown TLD:" in str(e):
    #         raise ValueError(f"Domain {domain} is not a valid domain")
    #     else:
    #         raise e

    # Parse response
    try:
        if domain_details is None:  # If the domain is not registered and query completely failed
            logging.info(f"""Exiting get_domain_info() and returning {{'creation_date': False,
                    'expiration_date': False,
                    'last_updated': False,
                    'name': {domain},
                    'name_servers': False,
                    'registrant_cc': False,
                    'registrar': False}}""")
            return {'creation_date': False,
                    'expiration_date': False,
                    'last_updated': False,
                    'name': domain,
                    'name_servers': False,
                    'registrant_cc': False,
                    'registrar': False}

        elif not domain_details:  # If the domain is not registered
            logging.info(f"""Exiting get_domain_info() and returning {{'creation_date': False,
                    'expiration_date': False,
                    'last_updated': False,
                    'name': {domain},
                    'name_servers': False,
                    'registrant_cc': False,
                    'registrar': False}}""")
            return {'creation_date': False,
                    'expiration_date': False,
                    'last_updated': False,
                    'name': domain,
                    'name_servers': False,
                    'registrant_cc': False,
                    'registrar': False}

        else:  # If there was domain info
            logging.info(f"Exiting get_domain_info() and returning {vars(domain_details)}")
            return vars(domain_details)

    except UnboundLocalError: # When the variable never gets assigned after a failure
        logging.info(f"""Exiting get_domain_info() and returning {{'creation_date': False,
                    'expiration_date': False,
                    'last_updated': False,
                    'name': {domain},
                    'name_servers': False,
                    'registrant_cc': False,
                    'registrar': False}}""")
        return {'creation_date': False,
                    'expiration_date': False,
                    'last_updated': False,
                    'name': domain,
                    'name_servers': False,
                    'registrant_cc': False,
                    'registrar': False}


def domain_availability(domain_query: dict) -> tuple:
    """Checks the availability of a domain

    Parameters
    ----------
    domain_query : dict
        The dictionary representation of a whois.Domain object

    Returns
    -------
    tuple[str, bool]
        Returns a tuple with a printable string about availability, 
        and a bool that's True if it's available and False if not

    Examples
    --------
    Getting details about an available domain
    ```
    from sws.domains import get_domain_info, domain_availability

    domain_details = get_domain_info('kieranwood.com')
    domain_availability(domain_details) # ('Domain available', True)
    ```

    Getting details of an unavailable domain
    ```
    from sws.domains import get_domain_info, domain_availability

    domain_details = get_domain_info('kieranwood.ca')
    domain_availability(domain_details) # ('Domain unavailable until 6/November/2020', False)
    ```
    """
    logging.info(f"Entering get_domain_info(domain_query={domain_query})")
    if not domain_query["expiration_date"] or domain_query["expiration_date"] < datetime.today():
        logging.info("Domain available, returning ('Domain available', True)")
        return "Domain available", True
    else:
        logging.info(f"Domain unavailable, returning ('Domain {domain_query['name']} unavailable until {domain_query['expiration_date'].day}/{month_name[domain_query['expiration_date'].month]}/{domain_query['expiration_date'].year}', False)")
        return f"Domain {domain_query['name']} unavailable until {domain_query['expiration_date'].day}/{month_name[domain_query['expiration_date'].month]}/{domain_query['expiration_date'].year}", False


def _install_whois():
    """Used to install whois binary if it isn't available"""
    logging.info("Entering _install_whois()")
    # Setting up default downloads folder based on OS
    if os.name == "nt":
        DOWNLOAD_FOLDER = f"{os.getenv('USERPROFILE')}\\Downloads"
        INSTALL_FOLDER = os.path.realpath(f"{os.getenv('USERPROFILE')}\\..\\..\\whois")
    else:  # PORT: For *Nix systems
        DOWNLOAD_FOLDER = f"{os.getenv('HOME')}/Downloads"
        INSTALL_FOLDER = f"{os.getenv('HOME')}/whois"
    if not os.path.exists(INSTALL_FOLDER):
        try:
            subprocess.Popen("whois")  # Check if binary is installed
        except FileNotFoundError:
            if os.name == "nt":  # Install windows version of whois
                logging.info(f"System is windows manually installing: DOWNLOAD_FOLDER = {DOWNLOAD_FOLDER}, INSTALL_FOLDER = {INSTALL_FOLDER}")
                logging.info(f"Downloading whois from https://download.sysinternals.com/files/WhoIs.zip and installing to {INSTALL_FOLDER}")
                build(ZIPResource("whois", "https://download.sysinternals.com/files/WhoIs.zip", overwrite_agreement=True))
                move(f"{DOWNLOAD_FOLDER}{os.sep}whois", INSTALL_FOLDER)
                logging.warning("Beginning adding whois to path variable")
                _add_to_path(INSTALL_FOLDER) # TODO: When new version of python-whois-extended releases remove this call
                print("Whois has been installed, restart script") # TODO: When new version of python-whois-extended releases remove this call
                sys.exit()
            else:  # Linux Installation
                logging.info(f"System is nix, installing with APT: DOWNLOAD_FOLDER = {DOWNLOAD_FOLDER}, INSTALL_FOLDER = {INSTALL_FOLDER}")
                build(APTResource("whois", "whois", overwrite_agreement=True))



"""Module that provides a class for doing downloads that feature:

- 4xx and 5xx error catching
- Progress bars for downloads
- Additional download metadata
- Easy printable debugging

Examples
--------
### Downloading image to current folder using original filename
```
from sws.downloads import Download

d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png') 
d.download() # Downloads to ./sws-banner.png
``` 

### Downloading image to downloads folder
```
import os
from sws.downloads import Download

d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png', download_path=os.path.join(os.getenv('USERPROFILE'),'Downloads'))
d.download() # Downloads to <downloads folder>/sws-banner.png
``` 

### Downloading image to current folder with custom name
```
from sws.downloads import Download

d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png' , filename='image.png') 
d.download() # Downloads to ./image.png
``` 

### Using Debug printing
```
from sws.downloads import Download
d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png')
print(d) '''Prints:
Details for image.png
size: 19kb
downloaded: False
url: https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png
download path: C:\\Users\\Kieran\\Desktop
'''

print(repr(d)) # Prints: Download for https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png to download image.png to C:\\Users\\Kieran\\Desktop and has not been downloaded yet
```
"""

import os                 # Used to validate paths
import re                 # Used to parse for filename(s)
import logging            # Used to log errors and debug info
from typing import Union  # Used to specify multi-type parameters

# Third Party Dependencies
import requests           # Used to download Files, and get file metadata
from tqdm import tqdm     # Used to create a progress bar for active downloads


class Download:
    """Class used to help download files

    Attributes
    ----------
    url: str
        The URL for the download

    size: Union[bool, int]
        The size of the file in kb

    download_path: Union[bool, str]
        The path to download the file to (the folder where the file will download to)

    filename: Union[bool, str]
        The name of the file to download, defaults to headers where available

    downloaded:bool = False
        Set to True when file has been downloaded, and keeps files from being redownloaded

    Notes
    -----
    - Download also has custom `__repr__()` and `__str__()` functions to help with debugging. Calling `print()` on an instance gives a readable multiline string, and using `repr()` on an instance gives a good debugging string that is single-line.
    - Each instance can only be used to download ONE file, once `self.downloaded` has been set to True (which happens in `download()`)you cannot use it to download again

    Examples
    --------
    ### Downloading image to current folder using original filename
    ```
    from sws.downloads import Download

    d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png') 
    d.download() # Downloads to ./sws-banner.png
    ``` 

    ### Downloading image to downloads folder
    ```
    import os
    from sws.downloads import Download

    d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png', download_path=os.path.join(os.getenv('USERPROFILE'),'Downloads'))
    d.download() # Downloads to <downloads folder>/sws-banner.png
    ``` 

    ### Downloading image to current folder with custom name
    ```
    from sws.downloads import Download

    d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png' , filename='image.png') 
    d.download() # Downloads to ./image.png
    ``` 

    ### Using Debug printing
    ```
    from sws.downloads import Download
    d = Download('https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png')
    print(d) '''Prints:
    Details for image.png
    size: 19kb
    downloaded: False
    url: https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png
    download path: C:\\Users\\Kieran\\Desktop
    '''

    print(repr(d)) # Prints: Download for https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png to download image.png to C:\\Users\\Kieran\\Desktop and has not been downloaded yet
    ```
    """   

    def __init__(self, url: str, download_path: Union[bool, str] = False, filename: Union[bool, str] = False):
        self.url = url
        self.downloaded = False
        self.size = self._get_size()

        # Setup filename variable
        if not filename:
            self._get_filename()
        else:
            self.filename = filename

        # Setup download_path variable
        if not download_path:
            self.download_path = os.path.realpath(".")
        else:
            if os.path.exists(download_path):
                self.download_path = download_path
            else:
                raise ValueError(f"Provided download path {download_path} does not exist")


    def _get_filename(self):
        """Check http headers for filename"""
        logging.info("No filename provided, checking headers")
        if "Content-Disposition" in requests.get(self.url).headers.keys():
            self.filename = re.findall("filename=(.+)", requests.get(self.url).headers["Content-Disposition"])[0]
            logging.info(f"Found filename {self.filename}")
        else:
            self.filename = self.url.split("/")[-1]
            logging.info(f"Could not find filename, falling back to {self.filename}")


    def _get_size(self) -> int:
        """Get the size of the file download

        Returns
        -------
        int
            The size of the download in KB
        """
        try:
            response = requests.get(self.url)
            if response.status_code//100 in [4,5]: # 4xx or 5xx status codes
                raise ValueError(f"{self.url} is not a valid download link and returned response code {response.status_code}")
            size = response.headers.get('content-length')
            if size:
                return int(size)//1024
            else:
                raise ValueError(f"{self.url} is not a valid download link")
        except ValueError as e:
            raise e # Reraise the error
        except Exception as e: # Have to do catchall and re-raise as ValueError because hundreds of exception types can be raised 
            raise ValueError(f"{self.url} is not a valid url, connection failed to establish with error {e}")


    def download(self):
        """Download a file from self.url"""
        file_path = os.path.realpath(f"{self.download_path}{os.sep}{self.filename}")
        if not self.downloaded: # If file is not downloaded
            logging.info("Starting binary download")

            # Setting up necessary download variables
            file_stream = requests.get(self.url, stream=True) # The open http request for the file
            chunk_size = 1024 # Setting the progress bar chunk size to measure in kb

            # Setting up the download progress bar
            progress_bar = tqdm(total=self.size, unit='iB', unit_scale=True)
            progress_bar.set_description(f"Download progress for {self.filename}")

            # Write the incoming data stream to a file and update progress bar as it downloads
            with open(file_path, 'wb') as download_file: 
                for chunk in file_stream.iter_content(chunk_size): 
                    if chunk:
                        progress_bar.update(len(chunk))
                        download_file.write(chunk)
            progress_bar.close()
            self.downloaded = True


    def __str__(self):
        return f"""Details for {self.filename}
    size: {self.size}kb
    downloaded: {self.downloaded}
    url: {self.url}
    download path: {self.download_path}"""


    def __repr__(self):
        return f"Download for {self.url} to download {self.filename} to {self.download_path} {'and has been downloaded' if self.downloaded else 'and has not been downloaded yet'}"


if __name__ == "__main__":
    d = Download("https://raw.githubusercontent.com/Descent098/sws/master/docs/img/sws-banner.png")
    print(d)
    print(str(d))



"""The primary entrypoint for the sws script.

This module's main function contains all the argument
parsing for the sws script.

Module Variabes
---------------
usage : str
    A variable that defines the argument parsing for the script in standard POSIX format.
"""

# Python Standard library
import sys                        # Used to check number of arguments
from pprint import pprint         # Used to pretty-print to command line

# External Dependencies
from docopt import docopt         # Used to parse CLI arguments
from sdu.autocomplete import *    # Used to generate bash autocomplete

# Internal Dependencies
from sws.domains import *         # Import all domains utilities
from sws.youtube import *         # Import all youtube utilities
from sws.ssl_utilities import *   # Import all ssl_utilties functions
from sws.redirects import trace   # Import trace from redirect utilities
from sws.dns_utilities import *   # Import all dns utilitites

usage = """Super Web Scripts; A command line interface, API, and set of scripts for web tasks

Usage:
    sws [-h] [-v]
    sws dns <domain>
    sws youtube <url> [<path>]
    sws ssl <hostname> [-e] [-c]
    sws redirects <url> [<ignored>]
    sws domains <domain> [-e] [-r] [-d] [-a]
    

Options:
    -h --help               Show this help message and exit
    -v --version            Show program's version number and exit
    -e --expiry             If specified will check the expiry of ssl cert/domain
    -c --cert               If specified will print the full details of the SSL cert
    -r --registrar          Tells you who the domain is registered through
    -d --details            If specified will show full domain details
    -a --available          Gives information on whether a specific domain is available
"""

command_list = [  # Used for autocompletion generation
    command("dns", []),
    command("youtube", []),
    command("ssl", ["-e", "--expiry", "-c", "--cert"]),
    command("redirects", []),
    command("domains", ["-e", "--expiry", "-r", "--registrar", "-d", "--details", "-a", "--available"]),
]


def main():
    """Primary entrypoint for the sws script."""
    args = docopt(usage, version="sws V0.2.0")  # Grab arguments for parsing

    if len(sys.argv) == 1:  # if no arguments are provided
        print(usage)
        if not os.name == "nt":  # Generate bash autocomplete
            autocomplete_file_text = generate_bash_autocomplete(command_list)
            try:
                with open("/etc/bash_completion.d/sws.sh", "w") as autocomplete_file:
                    autocomplete_file.write(autocomplete_file_text)
                print("Bash autocompletion file written to /etc/bash_completion.d/sws.sh \nPlease restart shell for autocomplete to update")
            except PermissionError:
                print("Unable to write bash autocompletion file for sws are you sudo?")
        sys.exit()

    if args["dns"]:
        dns_dict = get_dns_records(args['<domain>'], as_dict=True)
        print(dns_result_table(args['<domain>'], dns_dict))

    elif args["ssl"]:  # Begin parsing for ssl subcommand
        if args["--expiry"]:  # If -e or --expiry is specified
            print(f"SSL cert on domain {args['<hostname>']} Expires on: {check_ssl_expiry(args['<hostname>'])}")
        if args["--cert"]:  # If -c or --cert is specified
            pprint(get_ssl_cert(args['<hostname>']))
        if not (args["--expiry"] or args["--cert"]):
            print(usage)
            sys.exit()

    elif args["redirects"]:  # Begin parsing for redirects subcommand
        if args["<ignored>"]:
            if args["<ignored>"].startswith("["):
                args["<ignored>"] = list(args["<ignored>"])
        try:
            trace(args["<url>"], args["<ignored>"], print_result=True)
        except ValueError as e:
            print(e)

    elif args["youtube"]:
        download(args["<url>"], args["<path>"])

    elif args["domains"]:  # Begin parsing for ssl subcommand
        domain_details = get_domain_info(args["<domain>"])

        if args["--expiry"]:  # If -e or --expiry is specified
            expiry_date = domain_details["expiration_date"]
            print(f"Domain {args['<domain>']} set to expire on {expiry_date.strftime('%d-%b-%Y %H:%M:%S')}")
        if args["--registrar"]:
            if domain_details["registrar"]:
                print(f"{args['<domain>']} is registered through {domain_details['registrar']}")
            else:
                print(f"{args['<domain>']} is not registered")
        if args["--details"]:
            if not domain_details['creation_date']:
                print(f"Domain {args['<domain>']} was not registered")
            pprint(domain_details)
        if args["--available"]:
            print(f"Domain {args['<domain>']} is available" if domain_availability(domain_details)[1] else f"{domain_availability(domain_details)[0]}")

        if not (args["--expiry"] or args["--registrar"] or args["--details"] or args["--available"]):
            print(usage)
            sys.exit()

    else:
        print(usage)
        if not os.name == "nt":  # Generate bash autocomplete
            autocomplete_file_text = generate_bash_autocomplete(command_list)
            try:
                with open("/etc/bash_completion.d/sws.sh", "w") as autocomplete_file:
                    autocomplete_file.write(autocomplete_file_text)
                print("Bash autocompletion file written to /etc/bash_completion.d/sws.sh \nPlease restart shell for autocomplete to update")
            except PermissionError:
                print("Unable to write bash autocompletion file for sws are you sudo?")



"""A module for getting DNS configurations on a domain

Notes
-----
- If http or https is passed to any function it will be stripped

Examples
--------
### Printing the values of 'kieranwood.ca' when as_dict is True
```
from sws.dns_utilities import get_dns_records

print(get_dns_records("kieranwood.ca", as_dict=True)) # {'A': ['104.21.47.45', '172.67.144.116'], 'SOA': 'kevin.ns.cloudflare.com. dns.cloudflare.com. 2036568886 10000 2400 604800 3600'}
```
"""
# Standard Library Dependencies
import logging                              # Used for logging
from math import floor, ceil                # Used to normalize padding
from typing import Dict, List, Tuple, Union # Used to provide useful typehints in functions

# Third Party Library Dependencies
import dns.resolver     # Used to get domain information
import dns.rdatatype    # Used to determine record types

# Includes deprecated record types (more canonical)
RECORD_TYPES = dns.rdatatype.RdataType

# More human-readable and workable list that axes unused record types taken from https://en.wikipedia.org/wiki/List_of_DNS_record_types
RECORD_TYPES_TUPLE:Tuple[str] = ( 
    # Alias/resource records
    'A',
    'AAAA',
    'CNAME',
    'DNAME',
    'URI',
    'PTR',
    'LOC',
    'GPOS',

    # Nameservers/Name registrar config
    'NS',
    'SOA',
    'CSYNC',
    'NSEC',
    'NSEC3',
    'NSEC3PARAM',
    'CAA',
    'RP',
    'DHCID',

    # Mailservers/Communications
    'MX',
    'SPF',
    'SRV',
    'MD',  # Technically obsoleted by RFC 973
    'MF',  # Technically obsoleted by RFC 973

    #Mailing lists (mostly obsolete at this point but unfortunately not dead)
    'MB',
    'MG',
    'MR',
    'MAILA',
    'MAILB',
    'MINFO',

    # Cert info/configuration/misc
    'TXT',
    'CERT',
    'DNSKEY',
    'KEY',      # Somewhat obsoleted by RFC 3445
    'CDNSKEY',
    'HINFO',    # Unobsoleted by RFC 8482
    'AFSDB',
    'NAPTR',
    'KX',
    'OPT',
    'DS',
    'SSHFP',
    'IPSECKEY',
    'RRSIG',
    'TLSA',
    'HIP',
    'CDS',
    'EUI48',
    'EUI64',
    'TKEY',
    'TSIG',
    'IXFR',
    'AXFR',
    'TA',
    'DLV',
)


def get_dns_records(domain:str, as_dict:bool=False) -> Union[List[str], Dict[str, Union[str, List[str]]]]:
    """Takes in a domain and returns either a list or dictionary of the records of the domain

    Notes
    -----
    - A list of supported record types can be found by importing `sws.dns.RECORD_TYPES` or `sws.dns.RECORD_TYPES_TUPLE` (a tuple not an enum so you can do index lookups)
    - Records are returned asynchronously, this means the ordering can be off for records. For example if you have two NS records 
    `ns1.example.com` and `ns2.example.com` they could be put into the returned dict/list in either order 

    Parameters
    ----------
    domain : str
        [description]

    as_dict : bool, optional
        Whether to return a list of strings (False) or dictionary (True), by default False

    Returns
    -------
    Union[List[str], Dict[str, str]]
        If as_dict is False will return a list of printable strings (i.e. ['A: 127.0.0.1']), \n
        if as_dict is True will return a dict with record:str->record_data:str mapping (i.e. {'A':'127.0.0.1'})

    Raises
    ------
    ValueError
        If the domain has no valid records

    Examples
    --------
    ### Printing the values of 'kieranwood.ca'
    ```
    from sws.dns import get_dns_records

    print(get_dns_records("kieranwood.ca")) # ['A: 127.0.0.1', 'A: 106.23.49.45', 'NS: kevin.ns.cloudflare.com.']
    ```

    ### Printing the values of 'kieranwood.ca' when as_dict is True
    ```
    from sws.dns_utilities import get_dns_records

    print(get_dns_records("kieranwood.ca", as_dict=True)) # {'A': ['104.21.47.45', '172.67.144.116'], 'SOA': 'kevin.ns.cloudflare.com. dns.cloudflare.com. 2036568886 10000 2400 604800 3600'}
    ```
    """
    logging.info(f"Entering get_dns_records(domain={domain}, as_dict={as_dict}) ")

    if domain.startswith("https://"):
        logging.info(f"Stripping https:// protocol from {domain}")
        domain = domain.replace("https://", "")
    elif domain.startswith("http://"):
        logging.info(f"Stripping http:// protocol from {domain}")
        domain = domain.replace("http://", "")

    print(f"Beginning dns query to {domain} this may take up to 5 mins depending on connection speed")
    if as_dict:
        logging.debug("Beginning itteration of record types, and setting up result dictionary")
        result = {}
        for record_type in RECORD_TYPES:
            logging.info(f"Parsing record: {record_type}")
            try:
                try:
                    response = dns.resolver.resolve(domain, rdtype=record_type, lifetime=4.5)
                except dns.exception.Timeout:
                    ... # Record doesn't exist
                for record_data in response:
                    logging.info(f"Parsing record response {record_type}: {record_data.to_text()}")
                    if record_type == dns.rdatatype.RdataType.HTTPS:
                        result[record_type.name] = record_data.to_text().split(" ")[2:] # Remove 1 . from entries
                        continue
                    if result.get(record_type.name, False): # If multiple values of same record_type (i.e. 2+ A records on a domain)
                        logging.info(f"Found existing record data for {record_type}: {result.get(record_type.name, False)}")
                        if type(result.get(record_type.name, False)) == str: # If this is the second of same record_type
                            result[record_type.name] = [result[record_type.name], record_data.to_text()] # Convert string to list, and append new record data

                        elif type(result.get(record_type.name, False)) == list: # If 3+ of same record_type
                            logging.info(f"Found existing record data for {record_type}: {result.get(record_type.name, False)}")
                            result[record_type.name].append(record_data.to_text())

                    else: # If new record without existing value
                        logging.info(f"Writing data; {record_type}: {record_data.to_text()}")
                        result[record_type.name] = record_data.to_text()
            except Exception:
                ... # Record doesn't exist
    else:
        result = []
        for record_type in RECORD_TYPES:
            logging.info(f"Parsing record: {record_type}")
            try:
                try:
                    response = dns.resolver.resolve(domain, rdtype=record_type, lifetime=4.5)
                except dns.exception.Timeout:
                    ... # Record doesn't exist
                for record_data in response:
                    logging.info(f"Parsing record response {record_type}: {record_data.to_text()}")
                    result.append(f"{record_type.name}: {record_data.to_text()}")
            except Exception:
                ... # Record doesn't exist
    if not result:
        raise ValueError(f"Domain {domain} did not have any configured records, please check spelling")
    logging.info(f"Exiting get_dns_records() and returning {result}")
    return result


def dns_result_table(domain:str, dns_dict:dict) -> str:
    """Takes in a dictionary of dns values and returns a human-readable table

    Parameters
    ----------
    domain:str
        The domain used to generate the dns_dict

    dns_dict : dict
        A dictionary with all the dns records in a record_type(str)->record_value(str or list) mapping

    Returns
    -------
    str
        A table of dns records and their values for the domain
    """    
    logging.info(f"dns_result_table(domain={domain}, dns_dict={dns_dict}")
    if domain.startswith("https://"):
        logging.info(f"Stripping https:// protocol from {domain}")
        domain = domain.replace("https://", "")
    elif domain.startswith("http://"):
        logging.info(f"Stripping http:// protocol from {domain}")
        domain = domain.replace("http://", "")
    # add header
    result = f"""\nDNS records for {domain} \n
| Record Type | Record Value |
|-------------|--------------|\n"""
    logging.info("Starting record parsing")
    for record in dns_dict:
        logging.debug(f"Parsing record {record}: {dns_dict[record]}")

        # Add in Record Type column
        logging.debug(f"Creating record type column for {record} record")
        record_type = _even_padding(record, 13)
        current_row = f"|{record_type}|"

        # Add in Record Value column
        logging.debug(f"Creating record value column for {record} record")
        if type(dns_dict[record]) == list: # Need to render multiple result rows
            logging.debug("Current record has multiple values, therefore needs multiple rows")
            for index, record_value in enumerate(dns_dict[record]):
                if index == 0: # The first row
                    logging.debug(f"Creating first row for {record}:{record_value}")
                    value = _even_padding(record_value, 14)
                    current_row += f"{value}|\n"
                elif index == len(dns_dict[record])-1: # Last row for current record
                    logging.debug(f"Creating last row for {record}:{record_value}")
                    value = _even_padding(record_value, 14)
                    current_row += f"|{' '* 13}|{value}|\n|{'='* 13}|{'='* 14}|\n"

                else: # Middle row(s) for current record
                    logging.debug(f"Creating middle row for {record}:{record_value}")
                    value = _even_padding(record_value, 14)
                    current_row += f"|{' '* 13}|{value}|\n"

        else: # Just a single string for the current record so only takes a single row
            logging.debug("Current record has single value, generating column")
            value = _even_padding(dns_dict[record], 14)
            current_row += f"{value}|\n|{'='* 13}|{'='* 14}|\n"

        logging.debug(f"Appending record row(s) for {record} record to result string")
        result += current_row

    logging.info(f"Exiting dns_result_table() and returning {result}")
    return result


def _even_padding(value:str, size:int, spacer:str= " ") -> str:
    """Creates even padding for a string value within a set size

    Parameters
    ----------
    value : str
        The value to pad

    size : int
        The size (number of characters) total that are available

    spacer : str, optional
        What character to space with, by default " "

    Returns
    -------
    str
        The value with padding to make it even spaced within size,
        just returns original value if it's larger than size

    Examples
    --------
    ### Passing a value that is shorter or equal to size
    ```
    # Value same length as size
    value = '192.85.111.115'
    padded_value = even_padding(value, 14)
    print("|{padded_value}|") # Prints |192.85.111.115|
    
    # Value smaller than size
    value = '192.33.22.55'
    padded_value = even_padding(value, 14)
    print("|{padded_value}|") # Prints | 192.33.22.55 |
    ```

    ### Passing a value that larger than size
    ```
    value = 'karen.ns.cloudflare.com'
    padded_value = even_padding(value, 14)
    
    print("|{padded_value}|") # Never get's called cause value is too large
    ```
    """
    logging.debug(f"Entering _even_padding(value={value}, size={size}, spacer={spacer})")
    if len(value) > size:
        return value
    padding = abs(len(value) - size)/2
    if not padding.is_integer():
        padding_left = spacer * floor(padding)
        padding_right = spacer * ceil(padding)
    else:
        padding_left = spacer * int(padding)
        padding_right = spacer * int(padding)
    logging.debug(f"Exiting _even_padding() and returning:\n{padding_left}{value}{padding_right}")
    return f"{padding_left}{value}{padding_right}"


if __name__ == "__main__":
    result = get_dns_records("kieranwood.ca", as_dict=True)
    print(dns_result_table("kieranwood.ca", result))



"""An api and cli for doing common web development tasks

Goals
-----
This project has several goals:

1. Provide cross-platform implementations of functions (where possible)
2. Serve as a repository of general knowledge to share solutions to issues with people
3. Give people access to source code to learn how the functions work and just pull what's needed
4. Provide a useful API for well developed functions with error catching and testing for commonly annoying situations

Installation
------------
### From PyPi

run ```pip install sws``` or ```sudo pip3 install sws```.

### From source

1. Clone the github repo ([https://github.com/Descent098/sws](https://github.com/Descent098/sws))
2. cd into the 'sws' root directory (where setup.py is) and run ```pip install .``` or ```sudo pip3 install . ```

Modules
-------
### domains
A module for dealing with getting domain information such as:

- If a domain is available
- Who a domain is registered with
- Other domain details such as, creation_date, name_servers etc.

### dns_utilities
A module for getting DNS configurations on a domain

### downloads
Module that provides a class for doing downloads that feature:

- 4xx and 5xx error catching
- Progress bars for downloads
- Additional download metadata
- Easy printable debugging

### redirects
Provides a function for tracing redirects

### ssl_utilities
Get deails about the ssl cert of a hostname such as:

- When the cert will expire
- The issuer of the cert
- A full dict of the details of the cert

### youtube
Provides functionality for working with YouTube videos such as:

- Downloading youtube videos
- Pulling video metadata

Notes
-----
- When using the domain module if ```whois``` is not installed, the package/executable will be installed
- You should use an FQDN for any ```ssl_utilties``` or ```domains``` functions, so something like https://www.google.ca becomes google.ca
- All functions include logging and can be attached to for debugging assitance
"""


__pdoc__ = {"sws.cli": False} # Ignore the CLI in the API docs



"""Provides a function for tracing redirects

Examples
--------
### Trace the redirects from kieranwood.ca
```
from sws.redirects import trace

trace('kieranwood.ca', print_result = True) '''Prints:

Printing trace for http://kieranwood.ca

Redirect level:1
URL: http://kieranwood.ca/
HTTP Code: 301

Redirect level:2
URL: https://kieranwood.ca
HTTP Code: 200'''
```
"""

# Standard library Dependencies
import logging                  # Used for logging
from typing import Union        # Used for type hints with multiple types

# External Dependencies
import requests                 # Used to make http requests for redirect tracing


def trace(url: str, ignored_domains: Union[list, bool], print_result: bool = True) -> list:
    """Trace all redirects associated with a URL.

    Arguments
    ---------
    url : str
        The URL to trace, can include or not include a protocol

    ignored_domains : list[str] or bool
        A list of domains (without protocols) to ignore in the trace; False
        can be passed in if no domains should be ignored

    print_result : bool
        If true then the value will be printed in a human readable format

    Notes
    -----
    url argument can include or not include a protocol

    Raises
    ------
    ValueError:
        If url cannot be connected to a ValueError will be raised

    Returns
    -------
    list[responses]:
        The list of traced responses

    Examples
    --------
    Trace the redirects from kieranwood.ca
    ```
    from sws.redirects import trace

    trace('kieranwood.ca', print_result = True) '''Prints:

    Printing trace for http://kieranwood.ca

    Redirect level:1
    URL: http://kieranwood.ca/
    HTTP Code: 301

    Redirect level:2
    URL: https://kieranwood.ca
    HTTP Code: 200'''
    ```
    """
    logging.info(f"Entering trace(url={url}, ignored_domains={ignored_domains}, print_result={print_result})")

    logging.info(f"Checking protocol is present on {url}")
    # Checks if protocols are present
    if "https://" in url:
        ...  # Continue
    elif "http://" in url:
        ...  # Continue
    else:  # Add a protocol to URL
        url = "http://" + url
        logging.info(f"Changed url to {url}")

    # Try going to the provided URL
    logging.info("Starting HTTP request")
    try:
        response = requests.get(url)

    except requests.exceptions.ConnectionError:
        if print_result:
            print(f"Could not connect to {url}, please ensure there are no spelling mistakes")
        raise ValueError(f"Could not connect to {url}, please ensure there are no spelling mistakes")
    except Exception as identifier:
        if print_result:
            print(f"Error while checking {url} \nError Code: {identifier}")
        return [f"Error while checking {url} \nError Code: {identifier}"]

    output = []  # The result of the response
    if response.history:  # If the request was redirected
        if ignored_domains:
            logging.debug("Skipping ignored domains")
            response.history = _skip_ignored_domains(response.history, ignored_domains)
        if print_result:
            print(f"\nPrinting response for {url}")
        for level, redirect in enumerate(response.history):
            logging.debug(f"Appending redirect {redirect.url} to output")
            output.append([level+1, redirect.url, redirect.status_code])
        output.append([len(output)+1, response.url, response.status_code])
        if print_result:
            logging.debug("Printing result(s)")
            for redirect in output:
                print(f"\nRedirect level:{redirect[0]} \nURL: {redirect[1]} \nHTTP Code: {redirect[2]}")
        logging.info(f"Exiting trace() and returning {output}") 
        return output
    else:  # If the request was not redirected
        if print_result:
            print("Request was not redirected")
        logging.info("Exiting trace() and returning ['Request was not redirected']") 
        return ["Request was not redirected"]


def _skip_ignored_domains(response_trace: list, ignored_domains: list) -> list:
    """Takes a list of responses and removes any responses that
    have domains that are in the ignored_domains variable

    Arguments
    ---------
    response_trace : list[responses]
        List of responses to strip domain results from

    Notes
    -----
    ignored_domains argument can include or not include a protocol

    Returns
    -------
    list[responses]:
        The stripped list of responses

    Examples
    --------
    Skip all domains with safelinks.protection.outlook.com or can01.safelinks.protection.outlook.com in the responses
    ```
    from sws.utilities.redirects import trace

    trace('kieranwood.ca', ["safelinks.protection.outlook.com", "can01.safelinks.protection.outlook.com"], print_result = True)
    ```
    """
    # Remove instances of ignored domains from the response trace
    for domain in ignored_domains:
        for response in response_trace:
            if domain in response.url:
                response_trace.remove(response)
            else:
                continue
    return response_trace



"""Provides functionality for working with YouTube videos such as:

- Downloading youtube videos
- Pulling video metadata

Examples
--------
### Download youtube video to current folder
```
from sws.youtube import download

download("https://www.youtube.com/watch?v=6j1I3mC0BR0", ".") # Downloads video in current folder
```
"""

# Standard library Dependencies
import os                       # Used for path validation
import logging                  # Used in logging debug and info messages
import tkinter as tk            # Used to setup a gui for selecting files
from typing import Union        # Used for type hints with multiple types
from tkinter import filedialog  # Used to setup a gui for selecting files

# External Dependencies
from pytube import YouTube      # Used to download and get youtube video info
import pytube.exceptions


def _request_path():
    """Prompts user for the link to a youtube video, and path of where to download the video to"""
    logging.info("Entering _request_path()")
    try:
        root = tk.Tk()
        root.withdraw()
        file_path = str(filedialog.askdirectory(
            title="Select Video Output directory",
            mustexist=False))
    except Exception as e:
        print(f"Unable to start path selection GUI due to {e}, defualting downloading to current working directory {os.path.abspath('.')}")
        file_path = os.path.abspath(".")
    return file_path


def get_video_metadata(video_url:str) -> dict:
    """Gets the metadata of a youtube video

    Parameters
    ----------
    video_url : str
        The URL of the video

    Notes
    -----
    - Keys include:
        - title; The title of the video
        - id;  The id of the video
        - publish_date; The publishing date of the video (as a `datetime.datetime` object)
        - age_restricted; A boolean that is True if the video is age restricted
        - creator; A string representation of the channel name
        - metadata; A raw `pytube.metadata.YouTubeMetadata` object of video metadata

    Returns
    -------
    dict
        A dictionary of results (see notes for keys)

    Raises
    ------
    ValueError
        If the video is inaccessible for any reason
    """
    logging.info(f"Entering get_video_metadata(video_url={video_url})")
    result = {}
    try:
        logging.info("Getting video metadata")
        video_data = YouTube(video_url)
        result["title"] = video_data.title
        result["id"] = video_data.video_id
        result["publish_date"] = video_data.publish_date
        result["age_restricted"] = video_data.age_restricted
        result["creator"] = video_data.author
        result["metadata"] = video_data.metadata
    except pytube.exceptions.RegexMatchError:
        raise ValueError(f"Video URL {video_url} does not exist")
    except pytube.exceptions.VideoPrivate:
        raise ValueError(f"Video URL {video_url} is private")
    except pytube.exceptions.MembersOnly:
        raise ValueError(f"Video URL {video_url} is Members Only")
    except pytube.exceptions.RecordingUnavailable:
        raise ValueError(f"Video URL {video_url} does not have a recording for the livestream")
    except pytube.exceptions.VideoUnavailable:
        raise ValueError(f"Video URL {video_url} does not have an available video")
    except pytube.exceptions.LiveStreamError:
        raise ValueError(f"Video URL {video_url} is a livestream and cannot be downloaded")
    logging.info(f"Exiting get_video_metadata() and returning {result}")
    return result


def download(video_url: str, path: Union[str, bool]) -> str:
    """Downloads specified video_url to path

    Parameters
    ----------
    video_url : (str)
        The full URL you want to download i.e. https://www.youtube.com/watch?v=6j1I3mC0BR0

    path : (str or bool)
        The path on your system to download the video to, or False to start pathfinding function

    Raises
    ------
    ValueError:
        If the video is inaccessible for any reason

    Returns
    -------
    str:
        A string with 'Downloaded <video title> to <path> as <video_title>.mp4' if file was downloaded, or message 'File <filepath> already exists' if file already existed

    Examples
    --------
    Download youtube video to current folder
    ```
    from sws.youtube import download

    download("https://www.youtube.com/watch?v=6j1I3mC0BR0", ".") # Downloads video in current folder
    ```
    """
    logging.info(f"Entering download(video_url={video_url}, path={path})")
    if not path:
        logging.info("No path found, entering _request_path()")
        path = os.path.realpath(f"{_request_path()}")
    else:
        logging.info(f"Path found {path}, converting to abspath {os.path.abspath(path)}")
        path = os.path.abspath(path)

    # Get video title if video exists
    try:
        logging.info("Requesting video data")
        video_title = str(YouTube(video_url).title)
    except pytube.exceptions.RegexMatchError:
        raise ValueError(f"Video URL {video_url} does not exist")
    except pytube.exceptions.VideoPrivate:
        raise ValueError(f"Video URL {video_url} is private")
    except pytube.exceptions.MembersOnly:
        raise ValueError(f"Video URL {video_url} is Members Only")
    except pytube.exceptions.RecordingUnavailable:
        raise ValueError(f"Video URL {video_url} does not have a recording for the livestream")
    except pytube.exceptions.VideoUnavailable:
        raise ValueError(f"Video URL {video_url} does not have an available video")
    except pytube.exceptions.LiveStreamError:
        raise ValueError(f"Video URL {video_url} is a livestream and cannot be downloaded")

    logging.info("Checking if ouput file already exists")
    if os.path.exists(os.path.join(path, video_title)):
        logging.info(f"Found existing output file, exiting download() and returning 'File {os.path.join(path, video_title)} already exists'")
        return f"File {os.path.join(path, video_title)} already exists"

    print(f"Downloading {video_title} to {path}")
    YouTube(video_url).streams.get_highest_resolution().download(path)

    logging.info(f"Found existing output file, exiting download() and returning 'Downloaded {video_title} to {path} as {video_title}.mp4'")
    return f"Downloaded {video_title} to {path} as {video_title}.mp4"



"""Contains all the configuration for the package on pip"""
import setuptools

def get_content(*filename:str) -> str:
    """ Gets the content of a file or files and returns
    it/them as a string

    Parameters
    ----------
    filename : (str)
        Name of file or set of files to pull content from 
        (comma delimited)
    
    Returns
    -------
    str:
        Content from the file or files
    """
    content = ""
    for file in filename:
        with open(file, "r") as full_description:
            content += full_description.read()
    return content


setuptools.setup(
    name="sws",
    version="0.2.1",
    author="Kieran Wood",
    author_email="kieran@canadiancoding.ca",
    description="An API & command line interface, for common web tasks",
    long_description = get_content("README.md", "CHANGELOG.md"),
    long_description_content_type="text/markdown",
    project_urls={
        'API Docs': 'http://kieranwood.ca/sws',
        'CLI Docs': 'http://sws.readthedocs.io/',
        'Bug Reports': 'https://github.com/Descent098/sws/issues/new?assignees=Descent098&labels=bug&template=bug_report.md&title=%5BBUG%5D',
        'Source': 'https://github.com/Descent098/sws',
        'Roadmap': 'https://github.com/Descent098/sws/projects',
    },
    packages=setuptools.find_packages(),
    entry_points={
            'console_scripts': ['sws = sws.cli:main']
        },
    install_requires=[
    "requests",                 # Used in various modules for http connections and header parsing
    "pytube",                   # Used for youtube downloading
    "docopt",                   # Used for argument parsing in CLI
    "pystall",                  # Used to install ad-hoc binaries
    "python-whois-extended",    # Used to gather domain name info
    "dnspython",                # Used to gather dns info
    "locust",                   # Used to do loadtesting
    "sdu",                      # Allows for use of autocomplete on nix systems
    ],
    extras_require = {
        "dev" : ["nox",   # Used to run automated processes
                "pytest", # Used to run the test code in the tests directory
                "mkdocs", # Used to create HTML versions of the markdown docs in the docs directory
                ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta"
    ],
)